/**
 * Bundle of: axios-actions
 * Generated: 2018-11-26
 * Version: 3.0.0
 */

function isObject(value) {
    return typeof value === 'object' && !!value && !Array.isArray(value);
}
function isValue(value) {
    return typeof value !== 'undefined' && value !== null;
}
function getValue(obj, path) {
    var keys = path.split('.');
    while (obj && keys.length) {
        var key = keys.shift();
        obj = isObject(obj)
            ? obj[key]
            : undefined;
    }
    return obj;
}
/**
 * Flip the keys of an object
 *
 * @param   {object}  obj   A hash of key => value pairs
 */
function flipKeys(obj) {
    return Object.keys(obj).reduce(function (output, key) {
        output[obj[key]] = key;
        return output;
    }, {});
}
/**
 * Re-keys an object using a map of oldKey => newKey pairs
 *
 * @param {object}            obj   The source object to re-key
 * @param {object|function}   map   A map of keys or function
 * @param {boolean}           flip  A boolean to flip the mapping
 * @returns {*}
 */
function reKey(obj, map, flip) {
    if (map instanceof Function) {
        return map(obj, flip, reKey);
    }
    if (isObject(map)) {
        if (flip) {
            map = flipKeys(map);
        }
        return Object
            .keys(obj)
            .reduce(function (output, key) {
            output[map[key] || key] = obj[key];
            return output;
        }, {});
    }
    return obj;
}

function replaceTokens(template, data, pattern) {
    if ( pattern === void 0 ) pattern = /[:{](\w+[.\w]+)}?/;

    // replace array
    if (Array.isArray(data)) {
        var d = [].concat( data );
        while (d.length) {
            template = template.replace(pattern, d.shift());
        }
        return template;
    }
    // replace object
    var rx = new RegExp(pattern.source, 'g');
    return template.replace(rx, function (match, key) {
        var value = isObject(data)
            ? getValue(data, key)
            : data;
        return isValue(value)
            ? value
            : '';
    });
}

var Http = function Http(axios) {
    this.axios = axios;
    this.before = [];
    this.after = [];
    this.done = new Set();
    this.fail = new Set();
    this.queue = new Map;
    Object.freeze(this);
};
/**
 * Dispatch an axios request
 *
 * @param instance
 * @param config
 * @param data
 * @returns {Promise<any>}
 */
Http.prototype.request = function request (instance, config, data) {
        var this$1 = this;
        if ( data === void 0 ) data = null;

    // reset
    instance.error = null;
    instance.loading = true;
    // variables
    var method = config.method;
        var url = config.url;
    method = method.toLowerCase();
    // sanity check
    if (typeof this.axios[method] !== 'function') {
        throw new Error(("No such HTTP method '" + method + "'"));
    }
    // variables
    data = this.before.reduce(function (data, fn) { return fn(data); }, data);
    url = replaceTokens(url, data);
    // setup object
    config = Object.assign({}, config);
    config.method = method;
    config.data = data;
    config.url = url;
    // data
    if (method === 'get' && !url.includes('?') && isObject(data)) {
        config.params = data;
    }
    // loading
    var promise = this.axios.request(config);
    var key = Symbol((method + " " + url));
    this.queue.set(key, promise);
    var setLoaded = function (key) {
        this$1.queue.delete(key);
        instance.loading = this$1.queue.size > 0;
    };
    // call
    return promise
        .then(function (res) {
        setLoaded(key);
        this$1.after.forEach(function (fn) {
            var result = fn(res);
            if (typeof result !== 'undefined') {
                res = result;
            }
        });
        this$1.done.forEach(function (fn) { return fn(res); });
        return res;
    })
        .catch(function (error) {
        setLoaded(key);
        instance.error = error;
        this$1.fail.forEach(function (fn) { return fn(error); });
        return Promise.reject(error);
    });
};

/**
 * Helper function to process single object or array
 *
 * @param   data
 * @param   callback
 * @returns
 */
function process(data, callback) {
    if (Array.isArray(data)) {
        return data.map(function (obj) { return callback(obj); });
    }
    if (isObject(data)) {
        return callback(data);
    }
    return data;
}
/**
 * Default conversion function for resource plugin
 *
 * @param data
 * @returns {any}
 */
function toJSON(data) {
    return isObject(data) && data.toJSON instanceof Function
        ? data.toJSON()
        : data;
}

var helpers = /*#__PURE__*/Object.freeze({
  process: process,
  toJSON: toJSON
});

/**
 * Converts axios data to and from models
 *
 * @param api
 * @param model
 * @param convert
 */
function resource(api, model, convert) {
    if ( convert === void 0 ) convert = toJSON;

    api.http.before.push(function (data) { return process(data, function (data) { return convert(data); }); });
    api.http.after.push(function (res) { return process(res.data, function (data) { return new model(data); }); });
    return this;
}
/**
 * Remaps payload key names between client and server
 *
 * @param api
 * @param map
 */
function remap(api, map) {
    api.http.before.push(function (data) {
        process(data, function (obj) { return reKey(obj, map, false); });
        return data;
    });
    api.http.after.push(function (res) {
        res.data = process(res.data, function (obj) { return reKey(obj, map, true); });
        return res;
    });
    return this;
}
/**
 * Returns response data rather than the response itself
 *
 * @param api
 */
function data(api) {
    api.http.after.push(function (res) { return res.data; });
    return this;
}

var plugins = /*#__PURE__*/Object.freeze({
  resource: resource,
  remap: remap,
  data: data
});

function makeRequest(config, method, data) {
    if ( method === void 0 ) method = 'get';
    if ( data === void 0 ) data = null;

    var request = {};
    // request config was passed
    if (isObject(config)) {
        Object.assign(request, config);
    }
    // method / method + url passed
    else if (typeof config === 'string') {
        var url = config.trim();
        var matches = url.match(/^(get|post|patch|put|delete|head)\s+(.+)/i);
        if (matches) {
            method = matches[1];
            url = matches[2];
        }
        Object.assign(request, {
            method: method,
            url: url
        });
    }
    else {
        throw new Error('Invalid request parameters');
    }
    // cleanup
    request.method = String(request.method).toLowerCase();
    // return
    return request;
}

/**
 * Api class
 *
 * Base class to manage calls to API
 */
var ApiCore = function ApiCore(axios) {
    this.http = Object.freeze(new Http(axios));
    this.error = null;
    this.loading = false;
    Object.freeze(this.http);
};
/**
 * Use one of the built-in plugins
 *
 * @param   name  The name of the plugin
 * @param   paramsAny options the plugin needs
 * @returns {this}
 */
ApiCore.prototype.use = function use (name) {
        var params = [], len = arguments.length - 1;
        while ( len-- > 0 ) params[ len ] = arguments[ len + 1 ];

    var plugin = plugins[name];
    if (!plugin) {
        throw new Error(("No such plugin \"" + name + "\""));
    }
    plugin.apply(void 0, [ this ].concat( params ));
    return this;
};
/**
 * Query the API using a request config object
 *
 * @param   configThe HTTP method to make the call
 * @param   data  Any optional data to pass to the endpoints
 * @returns
 */
ApiCore.prototype.request = function request (config, data$$1) {
    return this.http.request(this, config, data$$1);
};
/**
 * Call the API via HTTP GET
 *
 * @param   url   The API URL to call
 * @param   data  Any optional data to pass to the endpoint
 * @returns
 */
ApiCore.prototype.get = function get (url, data$$1) {
    return this.request(makeRequest(url, 'get'), data$$1);
};
/**
 * Call the API via HTTP POST
 *
 * @param   url   The API URL to call
 * @param   data  Any optional data to pass to the endpoint
 * @returns
 */
ApiCore.prototype.post = function post (url, data$$1) {
    return this.request(makeRequest(url, 'post'), data$$1);
};
/**
 * Add a callback to fire when any call completes successfully
 *
 * @param   callback  The function to fire
 * @returns
 */
ApiCore.prototype.done = function done (callback) {
    this.http.done.add(callback);
    return this;
};
/**
 * Add a callback to fire when any call fails to complete
 *
 * @param   callback  The function to fire
 * @returns
 */
ApiCore.prototype.fail = function fail (callback) {
    this.http.fail.add(callback);
    return this;
};

var Action = function Action(config, handler) {
    this.config = config;
    if (handler) {
        this.addHandler(handler);
    }
};
Action.prototype.addHandler = function addHandler (handler) {
    if (!this.handlers) {
        this.handlers = new Set();
    }
    this.handlers.add(handler);
};
Action.prototype.removeHandler = function removeHandler (handler) {
    this.handlers.delete(handler);
};
Action.prototype.exec = function exec (data, action) {
    if (this.handlers) {
        this.handlers.forEach(function (handler) { return handler(data, action); });
    }
};

var ActionMap = function ActionMap(actions) {
    var this$1 = this;
    if ( actions === void 0 ) actions = null;

    this.map = {};
    if (actions) {
        Object
            .keys(actions)
            .forEach(function (name) {
            this$1.add(name, actions[name]);
        });
    }
};
ActionMap.prototype.get = function get (name) {
    return this.map[name];
};
/**
 * Add an action
 *
 * @param   name  The name of the action to add
 * @param   configAn object confirming to the AxiosRequestConfig interface
 * @param   handler   An optional handler function to fire on a successful call
 */
ActionMap.prototype.add = function add (name, config, handler) {
    this.map[name] = new Action(config, handler);
};
/**
 * Remove an action
 *
 * @param   name  The name of the action to delete
 */
ActionMap.prototype.remove = function remove (name) {
    delete this.map[name];
};
/**
 * Add callbacks to named actions
 *
 * @param   name  The name or names of actions to bind to
 * @param   handler   The function to fire
 * @returns this
 */
ActionMap.prototype.addHandler = function addHandler (name, handler) {
        var this$1 = this;

    var names = name.match(/\w+/g);
    names.forEach(function (name) {
        var action = this$1.get(name);
        if (action) {
            action.addHandler(handler);
        }
    });
    return this;
};
ActionMap.prototype.removeHandler = function removeHandler (name, handler) {
    var action = this.get(name);
    if (action) {
        action.removeHandler(handler);
    }
};

/**
 * Group class
 *
 * Manages related URLs of an API
 */
var ApiGroup = (function (ApiCore$$1) {
    function ApiGroup(axios, actions) {
        var this$1 = this;
        if ( actions === void 0 ) actions = null;

        ApiCore$$1.call(this, axios);
        this.actions = Object.freeze(new ActionMap());
        if (isObject(actions)) {
            Object
                .keys(actions)
                .forEach(function (action) {
                this$1.add(action, actions[action]);
            });
        }
    }

    if ( ApiCore$$1 ) ApiGroup.__proto__ = ApiCore$$1;
    ApiGroup.prototype = Object.create( ApiCore$$1 && ApiCore$$1.prototype );
    ApiGroup.prototype.constructor = ApiGroup;
    ApiGroup.prototype.add = function add (action, config, method, callback) {
        var this$1 = this;

        config = makeRequest(config, method);
        this.actions.add(action, config, callback);
        if (!(action in this)) {
            this[action] = function (data) {
                return this$1.call(action, data);
            };
        }
        return this;
    };
    /**
     * Call a specific action
     *
     * @param   name      The name of the action to execute
     * @param   data      An optional hash of data to pass to the server
     * @returns {Promise<any>}
     */
    ApiGroup.prototype.call = function call (name, data) {
        var action = this.actions.get(name);
        if (!action) {
            throw new Error(("No such action \"" + action + "\""));
        }
        return this
            .request(action.config, data)
            .then(function (res) {
            if (action.handlers) {
                action.exec(res, name);
            }
            return Promise.resolve(res);
        });
    };
    /**
     * Add a callback to fire when any call completes successfully
     *
     * @param   name      The name or names of actions to bind to
     * @param   callback  The function to fire
     * @returns this
     */
    ApiGroup.prototype.when = function when (name, callback) {
        this.actions.addHandler(name, callback);
        return this;
    };

    return ApiGroup;
}(ApiCore));

/**
 * Endpoint class
 *
 * Manages CRUD and index calls for a specific endpoint
 */
var ApiEndpoint = (function (ApiGroup$$1) {
    function ApiEndpoint(axios, config) {
        var this$1 = this;

        ApiGroup$$1.call(this, axios);
        // normal
        var actions = config;
        var methods = {
            read: 'get',
            index: 'get',
            create: 'post',
            update: 'post',
            delete: 'post'
        };
        // rest
        if (typeof config === 'string') {
            methods = {
                read: 'get',
                index: 'get',
                create: 'post',
                update: 'patch',
                delete: 'delete'
            };
            actions = Object
                .keys(methods)
                .reduce(function (output, action) {
                output[action] = config;
                return output;
            }, {});
        }
        // add actions
        Object
            .keys(actions)
            .map(function (action) {
            this$1.add(action, actions[action], methods[action]);
        });
    }

    if ( ApiGroup$$1 ) ApiEndpoint.__proto__ = ApiGroup$$1;
    ApiEndpoint.prototype = Object.create( ApiGroup$$1 && ApiGroup$$1.prototype );
    ApiEndpoint.prototype.constructor = ApiEndpoint;
    /**
     * Load the resource index
     * @param   data
     */
    ApiEndpoint.prototype.index = function index (data) {
        return this.call('index', data);
    };
    /**
     * Create a new resource
     * @param   data
     */
    ApiEndpoint.prototype.create = function create (data) {
        if (!isObject(data)) {
            throw new Error('Missing data parameter');
        }
        return this.call('create', data);
    };
    /**
     * Read a single resource
     * @param   id
     */
    ApiEndpoint.prototype.read = function read (id) {
        if (typeof id === 'undefined') {
            throw new Error('Missing id parameter');
        }
        return this.call('read', id);
    };
    /**
     * Update the resource
     * @param   data
     */
    ApiEndpoint.prototype.update = function update (data) {
        if (!isObject(data)) {
            throw new Error('Missing data parameter');
        }
        return this.call('update', data);
    };
    /**
     * Delete the resource
     * @param   id
     */
    ApiEndpoint.prototype.delete = function delete$1 (id) {
        if (typeof id === 'undefined') {
            throw new Error('Missing id parameter');
        }
        return this.call('delete', id);
    };

    return ApiEndpoint;
}(ApiGroup));

/**
 * Resource class
 *
 * Example class showing how Endpoint can be extended to be completely self-contained using plugins and handlers
 *
 * Featuring:
 *
 * - CRUD, index and search
 * - stores data on the class
 * - optional automatic model conversion
 * - optional auto-reload on create, update and delete
 */
var ApiResource = (function (ApiEndpoint$$1) {
    function ApiResource(axios, config, model, reload) {
        var this$1 = this;
        if ( reload === void 0 ) reload = false;

        ApiEndpoint$$1.call(this, axios, config);
        // data
        this.items = [];
        this.item = null;
        // add search
        var search = this.actions.get('search');
        if (!search) {
            var index = this.actions.get('index');
            this.add('search', index.config.url, index.config.method);
        }
        // always save data
        this
            .when('index search', function (items) {
            this$1.items = items;
        })
            .when('create read update delete', function (item) {
            this$1.item = item;
        });
        // optionally reload when updated
        if (reload) {
            this.when('create update delete', function (data) { return this$1.index(); });
        }
        // optionally use models
        if (model) {
            this.use('resource', model);
        }
        // resolve data
        this.use('data');
    }

    if ( ApiEndpoint$$1 ) ApiResource.__proto__ = ApiEndpoint$$1;
    ApiResource.prototype = Object.create( ApiEndpoint$$1 && ApiEndpoint$$1.prototype );
    ApiResource.prototype.constructor = ApiResource;
    ApiResource.prototype.search = function search (data) {
        return this.call('search', data);
    };

    return ApiResource;
}(ApiEndpoint));

export default ApiCore;
export { ApiCore, ApiGroup, ApiEndpoint, ApiResource, plugins, helpers };
